#include "Util-pkg.h"
#include "P3d-pkg.h"
#include "Localpath-pkg.h"
#include "Graphic-pkg.h"
#ifdef LIGHT_PLANNER
#include "ManipulationArmData.hpp"
#endif

#define DEBUG_LEVEL 0

/* Retourne le maximum de la norme infinie de la vitesse sur [0 tf] */
double get_max_norm_inf_speed(configPt q_i, configPt q_f,double tf,int index_dof){

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [START] get_max_norm_inf_speed\n");
    }

    double x0=q_i[index_dof];
    double y0=q_i[index_dof+1];
    double z0=q_i[index_dof+2];

    double xf=q_f[index_dof];
    double yf=q_f[index_dof+1];
    double zf=q_f[index_dof+2];

    double vx0=q_i[index_dof+6];
    double vy0=q_i[index_dof+7];
    double vz0=q_i[index_dof+8];

    double vxf=q_f[index_dof+6];
    double vyf=q_f[index_dof+7];
    double vzf=q_f[index_dof+8];

    double tf_4=tf*tf*tf*tf;
    double tf_5=tf_4*tf;
    double tf_6=tf_5*tf;
    double tf_7=tf_6*tf;

    double mnv=fabs(vx0);
    if(fabs(vxf)>mnv) mnv=fabs(vxf);
    if(fabs(vy0)>mnv) mnv=fabs(vy0);
    if(fabs(vyf)>mnv) mnv=fabs(vyf);
    if(fabs(vz0)>mnv) mnv=fabs(vz0);
    if(fabs(vzf)>mnv) mnv=fabs(vzf);

    double tr_x=(420*vx0)/tf_6+(420*vxf)/tf_6+(840*x0)/tf_7-(840*xf)/tf_7;
    double tr_y=(420*vy0)/tf_6+(420*vyf)/tf_6+(840*y0)/tf_7-(840*yf)/tf_7;
    double tr_z=(420*vz0)/tf_6+(420*vzf)/tf_6+(840*z0)/tf_7-(840*zf)/tf_7;

    if(fabs(tr_x)>EPS10){
        double tr=((240*vx0)/tf_5+(180*vxf)/tf_5+(420*x0)/tf_6-(420*xf)/tf_6)/tr_x;
        if(tr>0 && tr<tf){
            double tr_3=tr*tr*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double tr_6=tr_5*tr;
            double c4=-(35*x0-35*xf+tf*(20*vx0+15*vxf))/tf_4;
            double c5=(84*x0-84*xf+tf*(45*vx0+39*vxf))/tf_5;
            double c6=-(70*x0-70*xf+tf*(36*vx0+34*vxf))/tf_6;
            double c7=(20*x0-20*xf+tf*(10*vx0+10*vxf))/tf_7;
            double vtr=fabs(7*c7*tr_6+6*c6*tr_5+5*c5*tr_4+4*c4*tr_3+vx0);
            if(vtr>mnv){
                mnv=vtr;
            }
        }
    }

    if(fabs(tr_y)>EPS10){
        double tr=((240*vy0)/tf_5+(180*vyf)/tf_5+(420*y0)/tf_6-(420*yf)/tf_6)/tr_y;
        if(tr>0 && tr<tf){
            double tr_3=tr*tr*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double tr_6=tr_5*tr;
            double c4=-(35*y0-35*yf+tf*(20*vy0+15*vyf))/tf_4;
            double c5=(84*y0-84*yf+tf*(45*vy0+39*vyf))/tf_5;
            double c6=-(70*y0-70*yf+tf*(36*vy0+34*vyf))/tf_6;
            double c7=(20*y0-20*yf+tf*(10*vy0+10*vyf))/tf_7;
            double vtr=fabs(7*c7*tr_6+6*c6*tr_5+5*c5*tr_4+4*c4*tr_3+vy0);
            if(vtr>mnv){
                mnv=vtr;
            }
        }
    }

    if(fabs(tr_z)>EPS10){
        double tr=((240*vz0)/tf_5+(180*vzf)/tf_5+(420*z0)/tf_6-(420*zf)/tf_6)/tr_z;
        if(tr>0 && tr<tf){
            double tr_3=tr*tr*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double tr_6=tr_5*tr;
            double c4=-(35*z0-35*zf+tf*(20*vz0+15*vzf))/tf_4;
            double c5=(84*z0-84*zf+tf*(45*vz0+39*vzf))/tf_5;
            double c6=-(70*z0-70*zf+tf*(36*vz0+34*vzf))/tf_6;
            double c7=(20*z0-20*zf+tf*(10*vz0+10*vzf))/tf_7;
            double vtr=fabs(7*c7*tr_6+6*c6*tr_5+5*c5*tr_4+4*c4*tr_3+vz0);
            if(vtr>mnv){
                mnv=vtr;
            }
        }
    }

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [END] get_max_norm_inf_speed\n");
    }

    return mnv;
}

/* Retourne le maximum de la norme infinie de l'acceleration sur [0 tf] */
double get_max_norm_inf_acceleration(configPt q_i, configPt q_f, double tf,int index_dof){

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [START] get_max_norm_inf_acceleration\n");
    }

    double mna=0;

    double x0=q_i[index_dof];
    double y0=q_i[index_dof+1];
    double z0=q_i[index_dof+2];

    double xf=q_f[index_dof];
    double yf=q_f[index_dof+1];
    double zf=q_f[index_dof+2];

    double vx0=q_i[index_dof+6];
    double vy0=q_i[index_dof+7];
    double vz0=q_i[index_dof+8];

    double vxf=q_f[index_dof+6];
    double vyf=q_f[index_dof+7];
    double vzf=q_f[index_dof+8];

    double tf_4=tf*tf*tf*tf;
    double tf_5=tf_4*tf;
    double tf_6=tf_5*tf;
    double tf_7=tf_6*tf;
    double tf_10=tf_5*tf_5;
    double tf_11=tf_10*tf;
    double tf_12=tf_11*tf;

    double tr_x=(4200*vx0)/tf_6+(4200*vxf)/tf_6+(8400*x0)/tf_7-(8400*xf)/tf_7;
    double tr_y=(4200*vy0)/tf_6+(4200*vyf)/tf_6+(8400*y0)/tf_7-(8400*yf)/tf_7;
    double tr_z=(4200*vz0)/tf_6+(4200*vzf)/tf_6+(8400*z0)/tf_7-(8400*zf)/tf_7;

    if(fabs(tr_x)>EPS10){
        double a=(2220*vx0)/tf_5+(1980*vxf)/tf_5+(4200*x0)/tf_6-(4200*xf)/tf_6;
        double b=sqrt((3528000*(x0-xf)*(x0-xf))/tf_12+(896400*vx0*vx0+1735200*vx0*vxf+896400*vxf*vxf)/tf_10+(3528000*(vx0+vxf)*(x0-xf))/tf_11);
        double c4=-(35*x0-35*xf+tf*(20*vx0+15*vxf))/tf_4;
        double c5=(84*x0-84*xf+tf*(45*vx0+39*vxf))/tf_5;
        double c6=-(70*x0-70*xf+tf*(36*vx0+34*vxf))/tf_6;
        double c7=(20*x0-20*xf+tf*(10*vx0+10*vxf))/tf_7;
        double tr=(a+b)/tr_x;
        if(tr>0 && tr<tf){
            double tr_2=tr*tr;
            double tr_3=tr_2*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double atr=fabs(42*c7*tr_5+30*c6*tr_4+20*c5*tr_3+12*c4*tr_2);
            if(atr>mna){
                mna=atr;
            }
        }
        tr=(a-b)/tr_x;
        if(tr>0 && tr<tf){
            double tr_2=tr*tr;
            double tr_3=tr_2*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double atr=fabs(42*c7*tr_5+30*c6*tr_4+20*c5*tr_3+12*c4*tr_2);
            if(atr>mna){
                mna=atr;
            }
        }
    }

    if(fabs(tr_y)>EPS10){
        double a=(2220*vy0)/tf_5+(1980*vyf)/tf_5+(4200*y0)/tf_6-(4200*yf)/tf_6;
        double b=sqrt((3528000*(y0-yf)*(y0-yf))/tf_12+(896400*vy0*vy0+1735200*vy0*vyf+896400*vyf*vyf)/tf_10+(3528000*(vy0+vyf)*(y0-yf))/tf_11);
        double c4=-(35*y0-35*yf+tf*(20*vy0+15*vyf))/tf_4;
        double c5=(84*y0-84*yf+tf*(45*vy0+39*vyf))/tf_5;
        double c6=-(70*y0-70*yf+tf*(36*vy0+34*vyf))/tf_6;
        double c7=(20*y0-20*yf+tf*(10*vy0+10*vyf))/tf_7;
        double tr=(a+b)/tr_y;
        if(tr>0 && tr<tf){
            double tr_2=tr*tr;
            double tr_3=tr_2*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double atr=fabs(42*c7*tr_5+30*c6*tr_4+20*c5*tr_3+12*c4*tr_2);
            if(atr>mna){
                mna=atr;
            }
        }
        tr=(a-b)/tr_y;
        if(tr>0 && tr<tf){
            double tr_2=tr*tr;
            double tr_3=tr_2*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double atr=fabs(42*c7*tr_5+30*c6*tr_4+20*c5*tr_3+12*c4*tr_2);
            if(atr>mna){
                mna=atr;
            }
        }
    }

    if(fabs(tr_z)>EPS10){
        double a=(2220*vz0)/tf_5+(1980*vzf)/tf_5+(4200*z0)/tf_6-(4200*zf)/tf_6;
        double b=sqrt((3528000*(z0-zf)*(z0-zf))/tf_12+(896400*vz0*vz0+1735200*vz0*vzf+896400*vzf*vzf)/tf_10+(3528000*(vz0+vzf)*(z0-zf))/tf_11);
        double c4=-(35*z0-35*zf+tf*(20*vz0+15*vzf))/tf_4;
        double c5=(84*z0-84*zf+tf*(45*vz0+39*vzf))/tf_5;
        double c6=-(70*z0-70*zf+tf*(36*vz0+34*vzf))/tf_6;
        double c7=(20*z0-20*zf+tf*(10*vz0+10*vzf))/tf_7;
        double tr=(a+b)/tr_z;
        if(tr>0 && tr<tf){
            double tr_2=tr*tr;
            double tr_3=tr_2*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double atr=fabs(42*c7*tr_5+30*c6*tr_4+20*c5*tr_3+12*c4*tr_2);
            if(atr>mna){
                mna=atr;
            }
        }
        tr=(a-b)/tr_z;
        if(tr>0 && tr<tf){
            double tr_2=tr*tr;
            double tr_3=tr_2*tr;
            double tr_4=tr_3*tr;
            double tr_5=tr_4*tr;
            double atr=fabs(42*c7*tr_5+30*c6*tr_4+20*c5*tr_3+12*c4*tr_2);
            if(atr>mna){
                mna=atr;
            }
        }
    }

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [END] get_max_norm_inf_acceleration\n");
    }

    return mna;
}

/* Retourne le maximum de la norme infinie du jerk sur [0 tf] */
double get_max_norm_inf_jerk(configPt q_i, configPt q_f, double tf,int index_dof){

    double mnj=0;

    double x0=q_i[index_dof];
    double y0=q_i[index_dof+1];
    double z0=q_i[index_dof+2];

    double xf=q_f[index_dof];
    double yf=q_f[index_dof+1];
    double zf=q_f[index_dof+2];

    double vx0=q_i[index_dof+6];
    double vy0=q_i[index_dof+7];
    double vz0=q_i[index_dof+8];

    double vxf=q_f[index_dof+6];
    double vyf=q_f[index_dof+7];
    double vzf=q_f[index_dof+8];

    double tf_4=tf*tf*tf*tf;
    double tf_5=tf_4*tf;
    double tf_6=tf_5*tf;
    double tf_7=tf_6*tf;

    double c4,c5,c6,c7;

    for(int k=0;k<3;k++){

        int nbsol=0;
        double z[3];

        switch(k){
            case 0:
                c4=-(35*x0-35*xf+tf*(20*vx0+15*vxf))/tf_4;
                c5=(84*x0-84*xf+tf*(45*vx0+39*vxf))/tf_5;
                c6=-(70*x0-70*xf+tf*(36*vx0+34*vxf))/tf_6;
                c7=(20*x0-20*xf+tf*(10*vx0+10*vxf))/tf_7;
                break;
            case 1:
                c4=-(35*y0-35*yf+tf*(20*vy0+15*vyf))/tf_4;
                c5=(84*y0-84*yf+tf*(45*vy0+39*vyf))/tf_5;
                c6=-(70*y0-70*yf+tf*(36*vy0+34*vyf))/tf_6;
                c7=(20*y0-20*yf+tf*(10*vy0+10*vyf))/tf_7;
                break;
            case 2:
                c4=-(35*z0-35*zf+tf*(20*vz0+15*vzf))/tf_4;
                c5=(84*z0-84*zf+tf*(45*vz0+39*vzf))/tf_5;
                c6=-(70*z0-70*zf+tf*(36*vz0+34*vzf))/tf_6;
                c7=(20*z0-20*zf+tf*(10*vz0+10*vzf))/tf_7;
                break;
        }

        double d=24*c4;
        double c=120*c5;
        double b=360*c6;
        double a=840*c7;

        if(fabs(a)>EPS10){
             double p=-(b*b)/(3*a*a)+(c/a);
             double r=(b/(27*a))*(((2*b*b)/(a*a))-(9*c)/a)+(d/a);
             double delta=-(4*p*p*p+27*r*r);
             if(fabs(delta)<EPS10){
                if(fabs(p)<EPS10 || fabs(r)<EPS10){
                   z[0]=0;
                   nbsol=1;
                }
                else{
                   z[0]=3*r/p;
                   z[1]=-3*r/(2*p);
                   nbsol=2;
                }
             }
             else{
                if(delta<0){
                    double u=pow(((-r+sqrt(-delta/27))/2),1.0/3.0);
                    double w=pow(((-r-sqrt(-delta/27))/2),1.0/3.0);
                    z[0]=u+w;
                    nbsol=1;
                }
                else{
                    z[0]=2*sqrt(-p/3)*cos((1/3)*acos(-r/2*sqrt(27/(-pow(p,3.0)))));
                    z[1]=2*sqrt(-p/3)*cos((1/3)*acos(-r/2*sqrt(27/(-pow(p,3.0))))+2*3.14159/3);
                    z[2]=2*sqrt(-p/3)*cos((1/3)*acos(-r/2*sqrt(27/(-pow(p,3.0))))+4*3.14159/3);
                    nbsol=3;
                }
            }
            for(int i=0;i<nbsol;i++){
                z[i]=z[i]-b/(3*a);
            }
        }
        else{
            a=b;
            b=c;
            c=d;
            if(fabs(a)>EPS10){
                double delta=b*b-4*a*c;
                if(fabs(delta)<EPS10){
                    z[0]=-b/(2*a);
                    nbsol=1;
                }
                else{
                    if(delta>0){
                       z[0]=(-b-sqrt(delta))/(2*a);
                       z[1]=(-b+sqrt(delta))/(2*a);
                       nbsol=2;
                    }
                    else{
                       nbsol=0;
                    }
                }
            }
            else{
                a=b;
                b=c;
                if(fabs(a)>EPS10){
                    z[0]=-b/a;
                    nbsol=1;
                }
                else{
                    nbsol=0;
                }
            }
        }

        for(int i=0;i<nbsol;i++){
            if(z[i]>0 && z[i]<tf){
                double z_1=z[i];
                double z_2=z_1*z_1;
                double z_3=z_2*z_1;
                double z_4=z_3*z_1;
                double nje=fabs(210*c7*z_4+120*c6*z_3+60*c5*z_2+24*c4*z_1);
                if(nje>mnj) mnj=nje;
            }
        }

    }

    return mnj;
}

/* Recherche du tf optimal pour la vitesse */
double get_tf_opt_speed(p3d_rob *robotPt, configPt q_i, configPt q_f){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] get_tf_opt_speed\n");
    }

    plm_hermite_str hermite_paramPt = lm_get_hermite_lm_param(robotPt);

    double tmin,tmidle,tmax,mnv;
    int nb_it;

    tmax=hermite_paramPt->dx_search;
    tmin=0;
    nb_it=0;
    mnv=get_max_norm_inf_speed(q_i,q_f,tmax,hermite_paramPt->x_dof);
    while(mnv>hermite_paramPt->vmax && nb_it<hermite_paramPt->nb_it_max){
        tmin=tmax;
        tmax+=hermite_paramPt->dx_search;
        mnv=get_max_norm_inf_speed(q_i,q_f,tmax,hermite_paramPt->x_dof);
        nb_it++;
    }
    nb_it=0;
    while(fabs(tmin-tmax)>hermite_paramPt->epsi_err_x &&
          fabs(mnv-hermite_paramPt->vmax)>hermite_paramPt->epsi_err_y &&
          nb_it<hermite_paramPt->nb_it_max){
        tmidle=(tmax+tmin)/2;
        mnv=get_max_norm_inf_speed(q_i,q_f,tmidle,hermite_paramPt->x_dof);
        if(mnv>=hermite_paramPt->vmax){
            tmin=tmidle;
        }
        else{
            tmax=tmidle;
        }
        nb_it++;
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] get_tf_opt_speed\n");
    }

    return tmax;
}

/* Recherche du tf optimal pour l'acceleration */
double get_tf_opt_acceleration(p3d_rob *robotPt, configPt q_i, configPt q_f){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] get_tf_opt_acceleration\n");
    }

    plm_hermite_str hermite_paramPt = lm_get_hermite_lm_param(robotPt);

    double tmin,tmidle,tmax,mna;
    int nb_it;

    tmax=hermite_paramPt->dx_search;
    tmin=0;
    nb_it=0;
    mna=get_max_norm_inf_acceleration(q_i,q_f,tmax,hermite_paramPt->x_dof);
    while(mna>hermite_paramPt->amax && nb_it<hermite_paramPt->nb_it_max){
        tmin=tmax;
        tmax+=hermite_paramPt->dx_search;
        mna=get_max_norm_inf_acceleration(q_i,q_f,tmax,hermite_paramPt->x_dof);
        nb_it++;
    }
    nb_it=0;
    while(fabs(tmin-tmax)>hermite_paramPt->epsi_err_x &&
          fabs(mna-hermite_paramPt->amax)>hermite_paramPt->epsi_err_y &&
          nb_it<hermite_paramPt->nb_it_max){
        tmidle=(tmax+tmin)/2;
        mna=get_max_norm_inf_acceleration(q_i,q_f,tmidle,hermite_paramPt->x_dof);
        if(mna>=hermite_paramPt->amax){
            tmin=tmidle;
        }
        else{
            tmax=tmidle;
        }
        nb_it++;
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] get_tf_opt_acceleration\n");
    }

    return tmax;
}

/* Recherche du tf optimal pour le jerk */
double get_tf_opt_jerk(p3d_rob *robotPt, configPt q_i, configPt q_f){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] get_tf_opt_jerk\n");
    }

    plm_hermite_str hermite_paramPt = lm_get_hermite_lm_param(robotPt);

    double tmin,tmidle,tmax,mnj;
    int nb_it;

    tmax=hermite_paramPt->dx_search;
    tmin=0;
    nb_it=0;
    mnj=get_max_norm_inf_jerk(q_i,q_f,tmax,hermite_paramPt->x_dof);
    while(mnj>hermite_paramPt->jmax && nb_it<hermite_paramPt->nb_it_max){
        tmin=tmax;
        tmax+=hermite_paramPt->dx_search;
        mnj=get_max_norm_inf_jerk(q_i,q_f,tmax,hermite_paramPt->x_dof);
        nb_it++;
    }
    nb_it=0;
    while(fabs(tmin-tmax)>hermite_paramPt->epsi_err_x &&
          fabs(mnj-hermite_paramPt->jmax)>hermite_paramPt->epsi_err_y &&
          nb_it<hermite_paramPt->nb_it_max){
        tmidle=(tmax+tmin)/2;
        mnj=get_max_norm_inf_jerk(q_i,q_f,tmidle,hermite_paramPt->x_dof);
        if(mnj>=hermite_paramPt->jmax){
            tmin=tmidle;
        }
        else{
            tmax=tmidle;
        }
        nb_it++;
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] get_tf_opt_jerk\n");
    }

    return tmax;
}

/* Recherche du tf optimal */
double get_tf_opt(p3d_rob *robotPt, configPt q_i, configPt q_f){
    double tf_opt=get_tf_opt_speed(robotPt,q_i,q_f);
    double tf_opt_acceleration=get_tf_opt_acceleration(robotPt,q_i,q_f);
    double tf_opt_jerk=get_tf_opt_jerk(robotPt,q_i,q_f);
    if(tf_opt<tf_opt_acceleration){
        tf_opt=tf_opt_acceleration;
    }
    if(tf_opt<tf_opt_jerk){
        tf_opt=tf_opt_jerk;
    }
    return tf_opt;
}

/* allocation of a data structure specific to the hermite local method */
p3d_hermite_data * p3d_alloc_spec_hermite_localpath(p3d_rob *robotPt,configPt q_i, configPt q_f){

  if(DEBUG_LEVEL>0){
      printf("ALEX_DEBUG : [START] p3d_alloc_spec_hermite_local_path\n");
  }

  p3d_hermite_data * hermite_data;

  if ((hermite_data = MY_ALLOC(p3d_hermite_data,1)) == NULL)
    return NULL;

  hermite_data->q_init = q_i;
  hermite_data->q_end = q_f;
  double tf = get_tf_opt(robotPt,q_i,q_f);
  hermite_data->tf = tf;

  double x0,y0,z0,xf,yf,zf,vx0,vy0,vz0,vxf,vyf,vzf,tf_4,tf_5,tf_6,tf_7;

  lm_hermite_str * hermite_paramPt=lm_get_hermite_lm_param(robotPt);

  x0=q_i[hermite_paramPt->x_dof];
  y0=q_i[hermite_paramPt->y_dof];
  z0=q_i[hermite_paramPt->z_dof];
  xf=q_f[hermite_paramPt->x_dof];
  yf=q_f[hermite_paramPt->y_dof];
  zf=q_f[hermite_paramPt->z_dof];

  vx0=q_i[hermite_paramPt->vx_dof];
  vy0=q_i[hermite_paramPt->vy_dof];
  vz0=q_i[hermite_paramPt->vz_dof];
  vxf=q_f[hermite_paramPt->vx_dof];
  vyf=q_f[hermite_paramPt->vy_dof];
  vzf=q_f[hermite_paramPt->vz_dof];

  tf_4=tf*tf*tf*tf;
  tf_5=tf_4*tf;
  tf_6=tf_5*tf;
  tf_7=tf_6*tf;

  hermite_data->c4.x=-(35*x0-35*xf+tf*(20*vx0+15*vxf))/tf_4;
  hermite_data->c4.y=-(35*y0-35*yf+tf*(20*vy0+15*vyf))/tf_4;
  hermite_data->c4.z=-(35*z0-35*zf+tf*(20*vz0+15*vzf))/tf_4;

  hermite_data->c5.x=(84*x0-84*xf+tf*(45*vx0+39*vxf))/tf_5;
  hermite_data->c5.y=(84*y0-84*yf+tf*(45*vy0+39*vyf))/tf_5;
  hermite_data->c5.z=(84*z0-84*zf+tf*(45*vz0+39*vzf))/tf_5;

  hermite_data->c6.x=-(70*x0-70*xf+tf*(36*vx0+34*vxf))/tf_6;
  hermite_data->c6.y=-(70*y0-70*yf+tf*(36*vy0+34*vyf))/tf_6;
  hermite_data->c6.z=-(70*z0-70*zf+tf*(36*vz0+34*vzf))/tf_6;

  hermite_data->c7.x=(20*x0-20*xf+tf*(10*vx0+10*vxf))/tf_7;
  hermite_data->c7.y=(20*y0-20*yf+tf*(10*vy0+10*vyf))/tf_7;
  hermite_data->c7.z=(20*z0-20*zf+tf*(10*vz0+10*vzf))/tf_7;

  if(DEBUG_LEVEL>0){
      printf("ALEX_DEBUG : [END] p3d_alloc_spec_hermite_local_path\n");
  }

  return hermite_data;
}

/* allocation of local path of type hermite */
p3d_localpath * p3d_alloc_hermite_localpath(p3d_rob *robotPt,
                    configPt q_i, configPt q_f,
                    int lp_id,int is_valid){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_alloc_hermite_localpath\n");
    }

  p3d_localpath * localpathPt = NULL;

  if ((localpathPt = MY_ALLOC(p3d_localpath, 1)) == NULL)
    return NULL;

  /* allocation of the specific part */
  localpathPt->specific.hermite_data =
    p3d_alloc_spec_hermite_localpath(robotPt,q_i,q_f);

  if (localpathPt->specific.hermite_data == NULL){
    /* allocation failed free everything and return NULL*/
    MY_FREE(localpathPt, p3d_localpath, 1);
    return NULL;
  }
  /* Initialization of the generic part */
  /* fields */
  localpathPt->type_lp = HERMITE;
  localpathPt->valid = is_valid;
  localpathPt->lp_id = lp_id;
  localpathPt->prev_lp = NULL;
  localpathPt->next_lp = NULL;
  localpathPt->length_lp=localpathPt->specific.hermite_data->tf;

  /* methods associated to the local path */
  /* compute the length of the local path */
  localpathPt->length =
          (double (*)(p3d_rob*, p3d_localpath*))(p3d_hermite_dist);
  /* computes the configuration at given parameter along the path */
  localpathPt->config_at_param =
    (configPt (*)(p3d_rob*, p3d_localpath*,
          double))(p3d_hermite_config_at_param);
  localpathPt->config_at_distance =
    (configPt (*)(p3d_rob*, p3d_localpath*,
          double))(p3d_hermite_config_at_param);
  localpathPt->extract_sub_localpath =
      (p3d_localpath* (*)(p3d_rob*, p3d_localpath*,
            double, double))(p3d_extract_hermite);
  /* extract from a local path a sub local path starting from parameter
     u1 and ending at parameter u2 */
  localpathPt->extract_by_param =
      (p3d_localpath* (*)(p3d_rob*, p3d_localpath*,
            double, double))(p3d_extract_hermite);
  /* destroy the localpath */
    localpathPt->destroy =
      (void (*)(p3d_rob*, p3d_localpath*))(p3d_hermite_destroy);
  /*copy the local path */
  localpathPt->copy =
      (p3d_localpath* (*)(p3d_rob*,
            p3d_localpath*))(p3d_copy_hermite_localpath);
  /* compute the cost of a local path */
  localpathPt->cost =
    (double (*)(p3d_rob*, p3d_localpath*))(p3d_hermite_cost);
  /* function that simplifies the sequence of two local paths: valid
     only for RS curves */
  localpathPt->simplify =
    (p3d_localpath* (*)(p3d_rob*, p3d_localpath*, int*))(p3d_simplify_hermite);
  /* write the local path in a file */
  localpathPt->write =
    (int (*)(FILE *, p3d_rob*, p3d_localpath*))(p3d_write_hermite);
  /* from a configuration on a local path, this function computes an
     interval of parameter on the local path on which all the points
     of the robot move by less than the distance given as input.
     The interval is centered on the configuration given as input. The
     function returns the half length of the interval */
  localpathPt->stay_within_dist =
    (double (*)(p3d_rob*, p3d_localpath*,
        double, whichway, double*))(p3d_hermite_stay_within_dist);

  localpathPt->range_param = localpathPt->length_lp;
  localpathPt->ikSol = NULL;
  //save the active constraints
  localpathPt->nbActiveCntrts = 0;
  localpathPt->activeCntrts = NULL;

#ifdef LIGHT_PLANNER
  localpathPt->isCarryingObject = FALSE;
  for (int i = 0; i < MAX_CARRIED_OBJECTS; i++) {
    localpathPt->carriedObject[i] = NULL;
  }
#endif

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_alloc_hermite_localpath\n");
    }

  return localpathPt;
}

/* distance for the hermite local method
 * The output is actually the time of the motion between qi and qf
 * along the local path
 * */
double p3d_hermite_dist(p3d_rob *robotPt, p3d_localpath *localpathPt)
{

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_hermite_dist\n");
    }

  /* cast the pointer to union p3d_lm_specific to a pointer
     to p3d_hermite_data */
  p3d_hermite_data *specificPt = localpathPt->specific.hermite_data;

  /* test whether the type of local path is the expected one */
  if (localpathPt->type_lp != HERMITE){
    PrintError(("p3d_hermite_dist: Hermite local local path expected\n"));
  }

    if(specificPt==NULL && DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : specificPt is NULL in p3d_hermite_dist\n");
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_hermite_dist\n");
    }

  return specificPt->tf;
}

/*
 *  Compute the configuration situated at given parameter on the local path.
 *
 *  Input:  the robot, the distance.
 *
 *  Output: the configuration
 */
configPt p3d_hermite_config_at_param(p3d_rob *robotPt,
                    p3d_localpath *localpathPt,
                    double param)
{

    if(DEBUG_LEVEL!=0){
        printf("ALEX_DEBUG : [START] p3d_hermite_config_at_param\n");
    }

  p3d_hermite_data *hermite_specificPt;
  plm_hermite_str hermite_paramPt = lm_get_hermite_lm_param(robotPt);
  configPt q_init, q_end, q;
  double x,y,z,rx,ry,rz,vx,vy,vz,ax,ay,az;
  double Tx,Ty,Tz,T;
  double m,f,g;
  double c4x,c4y,c4z,c5x,c5y,c5z,c6x,c6y,c6z,c7x,c7y,c7z;
  double x0,y0,z0,vx0,vy0,vz0,rz0,rzf;
  double param_2,param_3,param_4,param_5,param_6,param_7;

  if (localpathPt == NULL)
    return NULL;

  if (localpathPt->type_lp != HERMITE){
    PrintError(("p3d_hermite_config_at_param : local path must be Hermite\n"));
    return NULL;
  }

  hermite_specificPt = localpathPt->specific.hermite_data;

  if (hermite_specificPt == NULL){
      return NULL;
  }

  m=hermite_paramPt->mass;
  g=hermite_paramPt->gravity;
  f=hermite_paramPt->friction;

  c4x=hermite_specificPt->c4.x;
  c4y=hermite_specificPt->c4.y;
  c4z=hermite_specificPt->c4.z;
  c5x=hermite_specificPt->c5.x;
  c5y=hermite_specificPt->c5.y;
  c5z=hermite_specificPt->c5.z;
  c6x=hermite_specificPt->c6.x;
  c6y=hermite_specificPt->c6.y;
  c6z=hermite_specificPt->c6.z;
  c7x=hermite_specificPt->c7.x;
  c7y=hermite_specificPt->c7.y;
  c7z=hermite_specificPt->c7.z;

  q_init = hermite_specificPt->q_init;
  q_end = hermite_specificPt->q_end;

  x0=q_init[hermite_paramPt->x_dof];
  y0=q_init[hermite_paramPt->y_dof];
  z0=q_init[hermite_paramPt->z_dof];

  vx0=q_init[hermite_paramPt->vx_dof];
  vy0=q_init[hermite_paramPt->vy_dof];
  vz0=q_init[hermite_paramPt->vz_dof];

  if(DEBUG_LEVEL<0){
      printf("V0=(%f,%f,%f)\n",vx0,vy0,vz0);
      printf("Vf=(%f,%f,%f)\n",q_end[hermite_paramPt->vx_dof],q_end[hermite_paramPt->vy_dof],q_end[hermite_paramPt->vz_dof]);
  }

  rz0=q_init[hermite_paramPt->rz_dof];
  rzf=q_end[hermite_paramPt->rz_dof];

  q = p3d_alloc_config(robotPt);

  if (param < 0)
    param = 0;
  if (param > localpathPt->length_lp)
    param = localpathPt->length_lp;
  if(localpathPt->length_lp == 0){
    p3d_destroy_config(robotPt, q);
    return p3d_copy_config(robotPt, q_init);
  }

  param_2=param*param;
  param_3=param_2*param;
  param_4=param_3*param;
  param_5=param_4*param;
  param_6=param_5*param;
  param_7=param_6*param;

  x=c7x*param_7+c6x*param_6+c5x*param_5+c4x*param_4+vx0*param+x0;
  y=c7y*param_7+c6y*param_6+c5y*param_5+c4y*param_4+vy0*param+y0;
  z=c7z*param_7+c6z*param_6+c5z*param_5+c4z*param_4+vz0*param+z0;

  vx=7*c7x*param_6+6*c6x*param_5+5*c5x*param_4+4*c4x*param_3+vx0;
  vy=7*c7y*param_6+6*c6y*param_5+5*c5y*param_4+4*c4y*param_3+vy0;
  vz=7*c7z*param_6+6*c6z*param_5+5*c5z*param_4+4*c4z*param_3+vz0;

  ax=42*c7x*param_5+30*c6x*param_4+20*c5x*param_3+12*c4x*param_2;
  ay=42*c7y*param_5+30*c6y*param_4+20*c5y*param_3+12*c4y*param_2;
  az=42*c7z*param_5+30*c6z*param_4+20*c5z*param_3+12*c4z*param_2;

  Tx=m*ax+f*vx;
  Ty=m*ay+f*vy;
  Tz=m*az+f*vz+m*g;
  T=sqrt(SQR(Tx)+SQR(Ty)+SQR(Tz));

  ry=asin(Tx/T);
  rx=asin(-Ty/(T*cos(ry)));
  rz=param*(rzf-rz0)/hermite_specificPt->tf+rz0;

  q[hermite_paramPt->x_dof]=x;
  q[hermite_paramPt->y_dof]=y;
  q[hermite_paramPt->z_dof]=z;
  q[hermite_paramPt->rx_dof]=rx;
  q[hermite_paramPt->ry_dof]=ry;
  q[hermite_paramPt->rz_dof]=rz;

  q[hermite_paramPt->vx_dof]=vx;
  q[hermite_paramPt->vy_dof]=vy;
  q[hermite_paramPt->vz_dof]=vz;

  q[hermite_paramPt->vx_dof+1]=ax;
  q[hermite_paramPt->vy_dof+2]=ay;
  q[hermite_paramPt->vz_dof+3]=az;

  for(int k=0;k<hermite_paramPt->x_dof;k++){
      q[k]=param*(q_end[k]-q_init[k])/hermite_specificPt->tf+q_init[k];
  }
  for(int k=hermite_paramPt->vz_dof+1;k<robotPt->nb_dof;k++){
      q[k]=param*(q_end[k]-q_init[k])/hermite_specificPt->tf+q_init[k];
  }

    if(DEBUG_LEVEL!=0){
        printf("ALEX_DEBUG : [END] p3d_hermite_config_at_param\n");
    }

  return q;
}

/*
 *  Extract from an Hermite local path the sub local path starting
 *  at parameter t1 and ending at parameter t2.
 *  The length of the extracted local path is computed
 *
 *  If t2 > length local path, return end of local path
 */
p3d_localpath *p3d_extract_hermite(p3d_rob *robotPt,
                p3d_localpath *localpathPt,
                double t1, double t2)
{

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_extract_hermite\n");
    }

  configPt q1, q2;
  p3d_localpath *sub_localpathPt;

  q1 = p3d_hermite_config_at_param(robotPt, localpathPt, t1);
  if (robotPt->cntrt_manager->cntrts != NULL) {
    p3d_set_and_update_this_robot_conf_multisol(robotPt, q1, NULL, 0, localpathPt->ikSol);
    p3d_get_robot_config_into(robotPt, &q1);
  }
  q2 = p3d_hermite_config_at_param(robotPt, localpathPt, t2);
  if (robotPt->cntrt_manager->cntrts != NULL) {
    p3d_set_and_update_this_robot_conf_multisol(robotPt, q2, NULL, 0, localpathPt->ikSol);
    p3d_get_robot_config_into(robotPt, &q2);
  }

  sub_localpathPt = p3d_alloc_hermite_localpath(robotPt, q1, q2, FORWARD,localpathPt->valid);

  p3d_copy_iksol(robotPt->cntrt_manager, localpathPt->ikSol, &(sub_localpathPt->ikSol));
  sub_localpathPt->nbActiveCntrts = localpathPt->nbActiveCntrts;
  sub_localpathPt->activeCntrts = MY_ALLOC(int, sub_localpathPt->nbActiveCntrts);
  for(int i = 0; i < sub_localpathPt->nbActiveCntrts; i++){
    sub_localpathPt->activeCntrts[i] = localpathPt->activeCntrts[i];
  }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_extract_hermite\n");
    }

  return sub_localpathPt;
}

/*
 *  lm_get_hermite_lm_param --
 *
 *  find the first occurrence of Hermite local method parameters.
 */

lm_hermite_str *lm_get_hermite_lm_param(p3d_rob* robotPt) {

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [START] lm_get_hermite_lm_param\n");
    }

  lm_list_param_str *list_paramPt = robotPt->local_method_params;
  lm_hermite_str *resultPt = NULL;
  while (list_paramPt!=NULL) {
    if (list_paramPt->lpl_type != HERMITE) {
        list_paramPt = list_paramPt->next;
    }
    else{
        resultPt = (plm_hermite_str)(list_paramPt->lm_param);

        if(DEBUG_LEVEL>1){
            printf("ALEX_DEBUG : [END] lm_get_hermite_lm_param\n");
        }

        return resultPt;
    }
  }

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [END (NULL)] lm_get_hermite_lm_param\n");
    }

  return resultPt;
}

/*
 *  lm_create_hermite
 *
 *  create an Hermite local method structure
 *
 *  Input: dtab[0,1,2,3,4] = vmax, amax, espi_err_x, epsi_err_y, dx_search
 *         itab[0,1] = nb_itmax, num_joint
 */

lm_hermite_str *lm_create_hermite(p3d_rob *robotPt, double *dtab,
    int *itab) {

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] lm_create_hermite\n");
    }

  lm_hermite_str* hermite_paramPt = NULL;

  if ((hermite_paramPt = MY_ALLOC(lm_hermite_str, 1)) == NULL) {
    printf("  lm_create_hermite : allocation failed\n");
    return (NULL);
  }

  if(dtab[0]<=0){
    printf("  lm_create_hermite : vmax is negative or null\n");
    return NULL;
  }
  hermite_paramPt->vmax = dtab[0];
  if(dtab[1]<=0){
    printf("  lm_create_hermite : amax is negative or null\n");
    return NULL;
  }
  hermite_paramPt->amax = dtab[1];
  if(dtab[2]<=0){
    printf("  lm_create_hermite : jmax is negative or null\n");
    return NULL;
  }
  hermite_paramPt->jmax = dtab[2];
  if(dtab[3]<=0){
    printf("  lm_create_hermite : epsi_err_x is negative or null\n");
    return NULL;
  }
  hermite_paramPt->epsi_err_x = dtab[3];
  if(dtab[4]<=0){
    printf("  lm_create_hermite : epsi_err_y is negative or null\n");
    return NULL;
  }
  hermite_paramPt->epsi_err_y = dtab[4];
  if(dtab[5]<=0){
    printf("  lm_create_hermite : dx_search is negative or null\n");
    return NULL;
  }
  hermite_paramPt->dx_search = dtab[5];
  if(dtab[6]<=0){
    printf("  lm_create_hermite : mass is negative or null\n");
    return NULL;
  }
  hermite_paramPt->mass = dtab[6];
  if(dtab[7]<=0){
    printf("  lm_create_hermite : friction is negative or null\n");
    return NULL;
  }
  hermite_paramPt->friction = dtab[7];
  if(dtab[8]<=0){
    printf("  lm_create_hermite : gravity is negative or null\n");
    return NULL;
  }
  hermite_paramPt->gravity = dtab[8];
  if(itab[0]<=0){
    printf("  lm_create_hermite : nb_it_max is negative or null\n");
    return NULL;
  }
  hermite_paramPt->nb_it_max = itab[0];

  if(itab[1]<0 || itab[1]>robotPt->njoints){
    printf("  lm_create_hermite : num_joint is out of range\n");
    return NULL;
  }
  hermite_paramPt->num_joint = itab[1];
  p3d_jnt* jnt=robotPt->joints[hermite_paramPt->num_joint];
  if(jnt->type!=P3D_FREEFLYER && jnt->type!=P3D_QUADROTOR){
    printf("  lm_create_hermite : joint of type P3D_FREEFLYER or P3D_QUADROTOR expected\n");
    return NULL;
  }

  hermite_paramPt->x_dof=jnt->index_dof;
  hermite_paramPt->y_dof=jnt->index_dof+1;
  hermite_paramPt->z_dof=jnt->index_dof+2;
  hermite_paramPt->rx_dof=jnt->index_dof+3;
  hermite_paramPt->ry_dof=jnt->index_dof+4;
  hermite_paramPt->rz_dof=jnt->index_dof+5;
  hermite_paramPt->vx_dof=jnt->index_dof+6;
  hermite_paramPt->vy_dof=jnt->index_dof+7;
  hermite_paramPt->vz_dof=jnt->index_dof+8;

  hermite_paramPt->voro=0;

  /*
  printf("robotPt->nb_dof=%d\n",robotPt->nb_dof);
  printf("robotPt->nb_user_dof=%d\n",robotPt->nb_user_dof);

  for(int k=0;k<=robotPt->njoints;k++){
      printf("\nk=%d\n",k);
      jnt=robotPt->joints[k];
      printf("jnt->name=%s\n",jnt->name);
      printf("jnt->index_dof=%d\n",jnt->index_dof);
      printf("jnt->index_user_dof=%d\n",jnt->index_user_dof);
      printf("jnt->dof_equiv_nbr=%d\n",jnt->dof_equiv_nbr);
      printf("jnt->user_dof_equiv_nbr=%d\n",jnt->user_dof_equiv_nbr);
      for(int l=0;l<jnt->dof_equiv_nbr;l++){
          printf("   jnt->dof_data[%d].is_user=%d\n",l,jnt->dof_data[l].is_user);
          printf("   jnt->dof_data[%d].vmin=%f\n",l,jnt->dof_data[l].vmin);
          printf("   jnt->dof_data[%d].vmax=%f\n",l,jnt->dof_data[l].vmax);
      }
  }

  printf("\n");
  print_config(robotPt,robotPt->ROBOT_POS);
  printf("\n");
  */

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] lm_create_hermite\n");
    }

  return(hermite_paramPt);
}

/*
 *  p3d_create_hermite_local_method --
 *
 *  create data-structure to store parameters of Hermite local method
 */
int p3d_create_hermite_local_method(double *dtab, int *itab) {

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_create_hermite_local_method\n");
    }

  p3d_rob *robotPt = (p3d_rob*)p3d_get_desc_curid(P3D_ROBOT);

  plm_hermite_str hermite_paramPt = lm_get_hermite_lm_param(robotPt);

  /* test that Hermite local method has not been already initialized
     for this robot */
  if (hermite_paramPt != NULL) {
    PrintWarning(("  hermite already initialized\n"));
    return FALSE;
  }

  robotPt->lpl_type = HERMITE;

  hermite_paramPt = lm_create_hermite(robotPt, dtab, itab);

  if (hermite_paramPt != NULL) {
    robotPt->local_method_params =
      lm_append_to_list(robotPt->local_method_params,
                        (void*)hermite_paramPt,
                        HERMITE);
  }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_create_hermite_local_method\n");
    }

  return TRUE;
}

/*
 * destroys a structure of type p3d_hermite_data
 */
void p3d_destroy_hermite_data(p3d_rob* robotPt, p3d_hermite_data* hermite_dataPt)
{
    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_destroy_hermite_data\n");
    }

  if (hermite_dataPt != NULL){
    if (hermite_dataPt->q_init != NULL){
      p3d_destroy_config(robotPt, hermite_dataPt->q_init);
    }
    if (hermite_dataPt->q_end != NULL){
      p3d_destroy_config(robotPt, hermite_dataPt->q_end);
    }
    MY_FREE(hermite_dataPt, p3d_hermite_data, 1);
  }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_destroy_hermite_data\n");
    }

}

/*
 * destroy an Hermite local path
 */
void p3d_hermite_destroy(p3d_rob* robotPt, p3d_localpath* localpathPt)
{

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_hermite_destroy\n");
    }

  if (localpathPt != NULL){

    /* test whether the type of local path is the expected one */
    if (localpathPt->type_lp != HERMITE){
      PrintError(("p3d_hermite_destroy: Hermite local path expected\n"));
    }
    /* destroy the specific part */
    if (localpathPt->specific.hermite_data != NULL){
      p3d_destroy_hermite_data(robotPt, localpathPt->specific.hermite_data);
    }
    localpathPt->next_lp = NULL;
    localpathPt->prev_lp = NULL;
    if (localpathPt->ikSol){
      p3d_destroy_specific_iksol(robotPt->cntrt_manager, localpathPt->ikSol);
      localpathPt->ikSol = NULL;
    }
    MY_FREE(localpathPt->activeCntrts, int, localpathPt->nbActiveCntrts);
    MY_FREE(localpathPt, p3d_localpath, 1);
  }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_hermite_destroy\n");
    }

}

/*
 *  Copy one local path.
 *
 *  Input:  the robot, the local path.
 *
 *  Output: the copied local path
 */

p3d_localpath *p3d_copy_hermite_localpath(p3d_rob* robotPt,
                      p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_copy_hermite_localpath\n");
    }

  configPt q_i, q_f;
  q_i = p3d_copy_config(robotPt, localpathPt->specific.hermite_data->q_init);
  q_f = p3d_copy_config(robotPt, localpathPt->specific.hermite_data->q_end);

  p3d_localpath* copy_localpathPt = p3d_alloc_hermite_localpath(robotPt,
          q_i,q_f,0,TRUE);

  /* update length and range of parameter */
  copy_localpathPt->length_lp = localpathPt->specific.hermite_data->tf;
  copy_localpathPt->range_param = localpathPt->specific.hermite_data->tf;
  p3d_copy_iksol(robotPt->cntrt_manager, localpathPt->ikSol, &(copy_localpathPt->ikSol));
  copy_localpathPt->nbActiveCntrts = localpathPt->nbActiveCntrts;
  copy_localpathPt->activeCntrts = MY_ALLOC(int, copy_localpathPt->nbActiveCntrts);
  for(int i = 0; i < copy_localpathPt->nbActiveCntrts; i++){
    copy_localpathPt->activeCntrts[i] = localpathPt->activeCntrts[i];
  }

#ifdef LIGHT_PLANNER
  copy_localpathPt->isCarryingObject = localpathPt->isCarryingObject;
  for (int i = 0; i < MAX_CARRIED_OBJECTS; i++) {
    copy_localpathPt->carriedObject[i] = localpathPt->carriedObject[i];
  }
#endif

    if(DEBUG_LEVEL>1){
        print_hermite_data(robotPt,copy_localpathPt->specific.hermite_data);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_copy_hermite_localpath\n");
    }

  return copy_localpathPt;
}

/*
 *  Cost of a local path
 *
 *  Input:  the local path
 *
 *  Output: the cost
 */
double p3d_hermite_cost(p3d_rob *robotPt, p3d_localpath *localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_hermite_cost\n");
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_hermite_cost\n");
    }

  return localpathPt->length_lp;
}

/*
 *  does nothing
 */
p3d_localpath *p3d_simplify_hermite(p3d_rob *robotPt,
                               p3d_localpath *localpathPt, int *need_colcheck){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_simplify_hermite\n");
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_simplify_hermite\n");
    }

  return localpathPt;
}

/*
 *  p3d_write_hermite --
 *
 *  write a localpath of type Hermite in a file
 *
 *  ARGS IN  : a file descriptor,
 *             a robot,
 *             a localpath
 *
 *  ARGS OUT : TRUE if success,
 *             FALSE if input local path is not an Hermite one.
 */
int p3d_write_hermite(FILE *file, p3d_rob* robotPt,
                         p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_write_hermite\n");
    }

  p3d_hermite_data *hermite_dataPt = NULL;

  if (localpathPt->type_lp != HERMITE) {
    return FALSE;
  }

  fprintf(file, "\n\np3d_add_localpath HERMITE\n");

  hermite_dataPt = (pp3d_hermite_data)localpathPt->specific.hermite_data;
  /* write each RS segment */
  fprintf(file, "conf_init");
  fprint_config_one_line(file, robotPt, hermite_dataPt->q_init);
  fprintf(file, "\nconf_end ");
  fprint_config_one_line(file, robotPt, hermite_dataPt->q_end);
  fprintf(file, "\nc4 \t%f \t%f \t%f",hermite_dataPt->c4.x
                                     ,hermite_dataPt->c4.y
                                     ,hermite_dataPt->c4.z);
  fprintf(file, "\nc5 \t%f \t%f \t%f",hermite_dataPt->c5.x
                                     ,hermite_dataPt->c5.y
                                     ,hermite_dataPt->c5.z);
  fprintf(file, "\nc6 \t%f \t%f \t%f",hermite_dataPt->c6.x
                                     ,hermite_dataPt->c6.y
                                     ,hermite_dataPt->c6.z);
  fprintf(file, "\nc7 \t%f \t%f \t%f",hermite_dataPt->c7.x
                                     ,hermite_dataPt->c7.y
                                     ,hermite_dataPt->c7.z);
  fprintf(file, "\ntf %f",hermite_dataPt->tf);
  fprintf(file, "\np3d_end_local_path\n");

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_write_hermite\n");
    }

  return TRUE;
}

/*  p3d_hermite_stay_within_dist
 *
 *  Input:  the robot,
 *          the local path,
 *          the parameter along the curve,
 *          the direction of motion
 *          the maximal distance moved by all the points of the
 *          robot
 *
 *  Output: length of the interval of parameter the robot can
 *          stay on without any body moving by more than the input distance
 *
 *  Description:
 *          From a configuration on a local path, this function
 *          computes an interval of parameter on the local path on
 *          which all the points of the robot move by less than the
 *          distance given as input.  The interval is centered on the
 *          configuration given as input. The function returns the
 *          half length of the interval
 */
double p3d_hermite_stay_within_dist(p3d_rob* robotPt,
                p3d_localpath* localpathPt,
                double parameter, whichway dir,
                double *distances){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_hermite_stay_within_dist\n");
    }

  p3d_hermite_data *hermite_localpathPt=NULL;
  int i, j, njnt = robotPt->njoints;
  p3d_jnt *cur_jntPt, *prev_jntPt;
  configPt q_max_param, q_param;
  double max_param, min_param;
  double range_param = localpathPt->length_lp;
  p3d_stay_within_dist_data * stay_within_dist_data;

  /* local path has to be of type Hermite */
  if (localpathPt->type_lp != HERMITE){
    PrintError(("p3d_hermite_stay_within_dist: Hermite local path expected\n"));
    return 0;
  }
  //on recupere les donnees du localpath
  hermite_localpathPt = localpathPt->specific.hermite_data;

  /* store the data to compute the maximal velocities at the
     joint for each body of the robot */
  stay_within_dist_data = MY_ALLOC(p3d_stay_within_dist_data, njnt+2); /** ALLOC **/
  p3d_init_stay_within_dist_data(stay_within_dist_data);

  if (dir == FORWARD){
    q_max_param = hermite_localpathPt->q_end;
    min_param = max_param = range_param - parameter;
  }
  else{
    q_max_param = hermite_localpathPt->q_init;
    min_param = max_param = parameter;
  }
  /* Get the current config to have the modifications of the constraints */
  /* Supose that q_init and q_goal respect constraints */
  q_param = p3d_get_robot_config(robotPt);  /** ALLOC **/
  //q_param = localpathPt->config_at_param(robotPt,localpathPt,parameter);

  /* computation of the bounds for the linear and angular
     velocities of each body */

  int minJnt = 0;
  for(i=0; i<=njnt; i++) {
    cur_jntPt = robotPt->joints[i];
    prev_jntPt = cur_jntPt->prev_jnt;

    /* j = index of the joint to which the current joint is attached */
    if (prev_jntPt==NULL)
      { j = -1; } /* environment */
    else
      { j = prev_jntPt->num; }
    double bakMinParam = min_param;

    p3d_jnt_stay_within_dist(&(stay_within_dist_data[j+1]), cur_jntPt,
                 &(stay_within_dist_data[i+1]), &(distances[i]),
                 q_param, q_max_param, max_param, &min_param);
    if (min_param < bakMinParam){
      minJnt = cur_jntPt->num;
    }
  /* Rem: stay_within_dist_data[0] is bound to the environment */
  }

  MY_FREE(stay_within_dist_data, p3d_stay_within_dist_data, njnt+2);  /** FREE **/
  p3d_destroy_config(robotPt, q_param);  /** FREE **/

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_hermite_stay_within_dist\n");
    }

  return min_param;
}

p3d_localpath *p3d_hermite_localplanner( p3d_rob *robotPt, configPt qi, configPt qf, int * ikSol){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_hermite_local_planner(with speed)\n");
    }

    configPt initconfPt, goalconfPt;
    p3d_localpath *localpathPt;

    /* on verifie que les configurations de depart et d'arrivee existent */
    if(qi == NULL){
        PrintInfo((("MP: p3d_hermite_localplanner: no start configuration...\n")));
        p3d_set_search_status(P3D_ILLEGAL_START);
        return(NULL);
    }
    if(qf == NULL){
        PrintInfo((("MP: p3d_hermite_localplanner: no goal configuration...\n")));
        p3d_set_search_status(P3D_ILLEGAL_GOAL);
        return(NULL);
    }

    lm_hermite_str * hermite_paramPt=lm_get_hermite_lm_param(robotPt);
    qi[hermite_paramPt->vz_dof+1]=0;
    qi[hermite_paramPt->vz_dof+2]=0;
    qi[hermite_paramPt->vz_dof+3]=0;
    qf[hermite_paramPt->vz_dof+1]=0;
    qf[hermite_paramPt->vz_dof+2]=0;
    qf[hermite_paramPt->vz_dof+3]=0;

    /* copy of qi into initconfPt and of qf into goalconfPt */
    initconfPt = p3d_copy_config(robotPt, qi);
    goalconfPt = p3d_copy_config(robotPt, qf);
    if(p3d_get_search_verbose()){
      PrintInfo(("MP: p3d_hermite_localplanner : "));
      PrintInfo(("qi=("));
      //print_config(robotPt, initconfPt);
      PrintInfo((") ; "));
      PrintInfo(("qf=("));
      //print_config(robotPt, goalconfPt);
      PrintInfo((")\n"));
    }

    /* If initconfPt == goalconfPt, free initconfPt and goalconfPt
     and return NULL */
    if(p3d_equal_config(robotPt,initconfPt, goalconfPt)) {
        PrintInfo((("MP: p3d_hermite_localplanner: q_init = q_goal!\n")));
        p3d_set_search_status(P3D_CONFIG_EQUAL);
        p3d_destroy_config(robotPt, initconfPt);
        p3d_destroy_config(robotPt, goalconfPt);
        return(NULL);
    }

    localpathPt = p3d_alloc_hermite_localpath(robotPt,initconfPt,goalconfPt,0,TRUE);

    p3d_set_search_status(P3D_SUCCESS);

    localpathPt->ikSol = ikSol;
    localpathPt->activeCntrts = p3d_getActiveCntrts(robotPt,&(localpathPt->nbActiveCntrts));

    #if defined(LIGHT_PLANNER)
      localpathPt->isCarryingObject = robotPt->isCarryingObject;
      for (int i = 0; i < MAX_CARRIED_OBJECTS; i++){
        if(i < robotPt->armManipulationData->size() ){
          localpathPt->carriedObject[i] = (*robotPt->armManipulationData)[i].getCarriedObject(); /*!< pointer to the carried object (obstacle environment or robot body) */
        }
      }
        //p3d_mat4Copy(robotPt->Tgrasp, localpathMg->Tgrasp);
    #endif

        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] p3d_hermite_local_planner(with speed)\n");
        }

      return(localpathPt);

}


void print_hermite_data(p3d_rob * robotPt,p3d_hermite_data* hermite_data){

    if(hermite_data==NULL){
        printf("-- HERMITE_DATA is NULL ----------\n");
    }
    else{
        printf("-- [START] HERMITE_DATA ----------\n");
        printf("   -- [START] Qi ----------\n");
        print_config(robotPt,hermite_data->q_init);
        printf("   -- [END] Qi ------------\n");
        printf("   -- [START] Qf ----------\n");
        print_config(robotPt,hermite_data->q_end);
        printf("   -- [END] Qf ------------\n");
        printf("tf=%f\n",hermite_data->tf);
        printf("c4=(%f,%f,%f)\n",hermite_data->c4.x,hermite_data->c4.y,hermite_data->c4.z);
        printf("c5=(%f,%f,%f)\n",hermite_data->c5.x,hermite_data->c5.y,hermite_data->c5.z);
        printf("c6=(%f,%f,%f)\n",hermite_data->c6.x,hermite_data->c6.y,hermite_data->c6.z);
        printf("c7=(%f,%f,%f)\n",hermite_data->c7.x,hermite_data->c7.y,hermite_data->c7.z);
        plm_hermite_str hermite_paramPt=lm_get_hermite_lm_param(robotPt);
        if(hermite_paramPt!=NULL){
            printf("vmax=%f\n",hermite_paramPt->vmax);
            printf("amax=%f\n",hermite_paramPt->amax);
            printf("epsi_err_x=%f\n",hermite_paramPt->epsi_err_x);
            printf("epsi_err_y=%f\n",hermite_paramPt->epsi_err_y);
            printf("dx_search=%f\n",hermite_paramPt->dx_search);
            printf("mass=%f\n",hermite_paramPt->mass);
            printf("friction=%f\n",hermite_paramPt->friction);
            printf("gravity=%f\n",hermite_paramPt->gravity);
            printf("nb_it_max=%d\n",hermite_paramPt->nb_it_max);
        }
        printf("-- [END] HERMITE_DATA -------------\n");
    }
}

void p3d_alex_debug(p3d_rob * robotPt){

    printf("\nALEX_DEBUG : writing in ");

    char dest[50];
    strcpy(dest,"/home/aboeuf/");
    strcat(dest,robotPt->name);
    strcat(dest,".m");

    printf("%s\n\n",dest);

    FILE * matlabout = fopen(dest,"w");
    FILE * debug = fopen("/home/aboeuf/debug.txt","w");
    fprintf(matlabout,"%c File automaticaly generated by move3d\n",'%');
    fprintf(matlabout,"%c List of the joints and their positions\n\n",'%');
    fprintf(matlabout,"clear all\n");
    fprintf(matlabout,"close all\n");
    fprintf(matlabout,"figure\n");
    fprintf(matlabout,"hold on\n\n");
    int N=robotPt->njoints+1;
    fprintf(matlabout,"N=%d;\n\n",N);
    fprintf(matlabout,"scale=0.1;\n\n");

    fprintf(matlabout,"JOINTS_POS=[...\n");
    for(int k=0;k<=robotPt->njoints;k++){
        p3d_jnt* joint_k = robotPt->joints[k];
        p3d_point rel_p0=joint_k->relative_p0;
        fprintf(matlabout,"... %c [J%d]\n",'%',k);
        fprintf(matlabout,"%f %f %f %f\n",joint_k->pos0[0][0],joint_k->pos0[0][1],joint_k->pos0[0][2],joint_k->pos0[0][3]);
        fprintf(matlabout,"%f %f %f %f\n",joint_k->pos0[1][0],joint_k->pos0[1][1],joint_k->pos0[1][2],joint_k->pos0[1][3]);
        fprintf(matlabout,"%f %f %f %f\n",joint_k->pos0[2][0],joint_k->pos0[2][1],joint_k->pos0[2][2],joint_k->pos0[2][3]);
        fprintf(matlabout,"%f %f %f %f\n",joint_k->pos0[3][0],joint_k->pos0[3][1],joint_k->pos0[3][2],joint_k->pos0[3][3]);
    }
    fprintf(matlabout,"];\n\n");

    fprintf(matlabout,"PREV=[");
    for(int k=0;k<=robotPt->njoints;k++){
        if(robotPt->joints[k]->prev_jnt!=NULL){
            fprintf(matlabout,"%d ",robotPt->joints[k]->prev_jnt->num);
        }
        else{
            fprintf(matlabout,"-1 ");
        }
    }

    for(int k=0;k<=robotPt->njoints;k++){
        p3d_jnt* joint_k = robotPt->joints[k];
        fprintf(debug,"[%d] -> %s\n",joint_k->num,joint_k->name);
        fprintf(debug,"dof_equiv_number = %d\n",joint_k->dof_equiv_nbr);
        fprintf(debug,"[index_dof,index_user_dof] = [%d,%d]\n",joint_k->index_dof,joint_k->index_user_dof);

        fprintf(debug,"\n");
    }

    fprintf(matlabout,"];\n\n");

    fprintf(matlabout,"B0_prev=JOINTS_POS(1:3,4);\n");
    fprintf(matlabout,"for k=0:N-1\n");
    fprintf(matlabout,"   B0=JOINTS_POS(4*k+1:4*k+3,4);\n");
    fprintf(matlabout,"   B1=B0+scale*JOINTS_POS(4*k+1:4*k+3,1);\n");
    fprintf(matlabout,"   B2=B0+scale*JOINTS_POS(4*k+1:4*k+3,2);\n");
    fprintf(matlabout,"   B3=B0+scale*JOINTS_POS(4*k+1:4*k+3,3);\n");
    fprintf(matlabout,"   plot3(B0(1),B0(2),B0(3),'ko');\n");
    fprintf(matlabout,"   plot3([B0(1) B1(1)],[B0(2) B1(2)],[B0(3) B1(3)],'r');\n");
    fprintf(matlabout,"   plot3([B0(1) B2(1)],[B0(2) B2(2)],[B0(3) B2(3)],'g');\n");
    fprintf(matlabout,"   plot3([B0(1) B3(1)],[B0(2) B3(2)],[B0(3) B3(3)],'b');\n");
    fprintf(matlabout,"   if (PREV(k+1)~=-1)\n");
    fprintf(matlabout,"      B0_prev=JOINTS_POS(4*PREV(k+1)+1:4*PREV(k+1)+3,4);\n");
    fprintf(matlabout,"   end\n");
    fprintf(matlabout,"   plot3([B0(1) B0_prev(1)],[B0(2) B0_prev(2)],[B0(3) B0_prev(3)],'k--');\n");
    fprintf(matlabout,"end\n\n");

    fprintf(matlabout,"axis equal\n");

    fclose(matlabout);
    fclose(debug);

    print_config(robotPt,robotPt->ROBOT_POS);

}
